/*
 * Kivio - Visual Modelling and Flowcharting
 * Copyright (C) 2000 theKompany.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/**************************************************************************************
 *
 * The code for dragging and resizing stencils is all contained in this class. KivioCanvas
 * is used only for drawing since it's a canvas.
 *
 */

#include "tool_select.h"

#include "kivio_view.h"
#include "kivio_canvas.h"
#include "kivio_page.h"

#include "kivio_custom_drag_data.h"
#include "kivio_layer.h"
#include "kivio_point.h"
#include "kivio_stencil.h"
#include <X11/Xlib.h>

SelectTool::SelectTool( KivioView* view )
:Tool(view)
{
  setSortNum(0);
  controller()->setDefaultTool(this);

  ToolSelectAction* select = new ToolSelectAction( actionCollection(), "ToolAction" );
  KAction* m_z1 = new KAction( i18n("Select"), "kivio_arrow", 0, actionCollection(), "select" );
  select->insert(m_z1);

  m_mode = stmNone;
  m_pResizingStencil = NULL;
  m_pCustomDraggingStencil = NULL;

  m_customDragID = 0;
}

SelectTool::~SelectTool()
{
}


/**
 * Event delegation
 *
 * @param e The event to be identified and processed
 *
 */
void SelectTool::processEvent( QEvent* e )
{
    switch (e->type())
    {
    case QEvent::MouseButtonPress:
        mousePress( (QMouseEvent*)e );
        break;

    case QEvent::MouseButtonRelease:
        mouseRelease( (QMouseEvent*)e );
        break;

    case QEvent::MouseMove:
        mouseMove( (QMouseEvent*)e );
        break;

    default:
      break;
    }
}

void SelectTool::activate()
{
    debug("SelectTool activate");
    m_pCanvas->setCursor(arrowCursor);
    m_mode = stmNone;
}

void SelectTool::deactivate()
{
}

void SelectTool::configure()
{
}


/**
 * Selects all stencils inside a given rect
 */
void SelectTool::select(QRect r)
{
    // Calculate the start and end clicks in terms of page coordinates
    TKPoint startPoint = m_pCanvas->mapFromScreen( QPoint( r.x(), r.y() ) );
    TKPoint releasePoint = m_pCanvas->mapFromScreen( QPoint( r.x() + r.width(), r.y() + r.height() ) );


    float x, y, w, h;

    // Calculate the x,y position of the selection box
    x = startPoint.x < releasePoint.x ? startPoint.x : releasePoint.x;
    y = startPoint.y < releasePoint.y ? startPoint.y : releasePoint.y;

    // Calculate the w/h of the selection box
    w = releasePoint.x - startPoint.x;
    if( w < 0.0f )
        w *= -1.0f;

    h = releasePoint.y - startPoint.y;
    if( h < 0.0f )
        h *= -1.0f;

    // Tell the page to select all stencils in this box
    m_pCanvas->activePage()->selectStencils( x, y, w, h );
}

void SelectTool::mousePress( QMouseEvent *e )
{
    // Gets the list of keys held down and check if the shift key is one of them. If yes, set the flag
    XQueryKeymap( qt_xdisplay(), m_keys );
    if( !(m_keys[6] & 4 ) && !(m_keys[7] & 64) )
        m_shiftKey = false;
    else
        m_shiftKey = true;

    // Last point is used for undrawing at the last position and calculating the distance the mouse has moved
    m_lastPoint = m_pCanvas->mapFromScreen( e->pos() );

    // Check if we nailed a custom drag point
    if( startCustomDragging( e )==true )
    {
        m_mode = stmCustomDragging;
        return;
    }

    // Check if we are resizing
    if( startResizing( e )==true )
    {
        m_mode = stmResizing;
        return;
    }

    // Check if we can drag a stencil
    if( startDragging( e )==true )
    {
        m_mode = stmDragging;
        return;
    }

    // This should always be the last 'start' call since it always returns true
    if( startRubberBanding( e ) )
    {
        m_mode = stmDrawRubber;
        return;
    }
}


/**
 * Tests if we should start rubber banding (always returns true).
 */
bool SelectTool::startRubberBanding( QMouseEvent *e )
{
    // We didn't find a stencil, so unselect everything if we aren't holding the shift key down
    if( !m_shiftKey )
        m_pCanvas->activePage()->unselectAllStencils();

    m_pCanvas->startRectDraw( e->pos(), KivioCanvas::Rubber );
    m_pCanvas->repaint();

    return true;
}


/**
 * Tests if we can start dragging a stencil.
 */
bool SelectTool::startDragging( QMouseEvent *e )
{
    KivioPage *pPage = m_pCanvas->activePage();
    KivioPoint kPoint;
    KivioStencil *pStencil;
    int colType;

    TKPoint pagePoint = m_pCanvas->mapFromScreen( e->pos() );
    kPoint.set( pagePoint.x, pagePoint.y );

    pStencil = pPage->checkForStencil( &kPoint, &colType );

    if( !pStencil )
        return false;

    if( pStencil->isSelected() )
    {
        // If we are clicking an already selected stencil, and the shift
        // key down, then unselect this stencil
        if( m_shiftKey==true )
            pPage->unselectStencil( pStencil );

        // Otherwise, it means we are just moving
    }
    else
    {
        // Clicking a new stencil, and the shift key is not down
        if( !m_shiftKey )
            pPage->unselectAllStencils();

        pPage->selectStencil( pStencil );
    }

    // Set the mode
    m_mode = stmDragging;

    // Create a new painter object
    m_pCanvas->beginUnclippedSpawnerPainter();
    m_pCanvas->drawSelectedStencilsXOR();

    // Tell the view to update the toolbars to reflect the
    // first selected stencil's settings
    m_pView->updateToolBars();

    return true;
}

bool SelectTool::startCustomDragging( QMouseEvent *e )
{
    KivioPage *pPage = m_pCanvas->activePage();
    KivioPoint kPoint;
    KivioStencil *pStencil;
    int colType;

    TKPoint pagePoint = m_pCanvas->mapFromScreen( e->pos() );
    kPoint.set( pagePoint.x, pagePoint.y );

    pStencil = pPage->checkForStencil( &kPoint, &colType );

    if( !pStencil || colType < kctCustom )
        return false;


    if( pStencil->isSelected() )
    {
        // If we are clicking an already selected stencil, and the shift
        // key down, then unselect this stencil
        if( m_shiftKey==true )
        {
            m_pCustomDraggingStencil = NULL;
            pPage->unselectStencil( pStencil );
        }
        else
            m_pCustomDraggingStencil = pStencil;

        // Otherwise, it means we are just moving
    }
    else
    {
        // Clicking a new stencil, and the shift key is not down
        if( !m_shiftKey )
            pPage->unselectAllStencils();

        m_pCustomDraggingStencil = pStencil;
        
        pPage->selectStencil( pStencil );
    }

    // Set the mode
    m_mode = stmCustomDragging;

    m_customDragID = colType;

    // Create a new painter object
    m_pCanvas->beginUnclippedSpawnerPainter();
    m_pCanvas->drawSelectedStencilsXOR();

    return true;
}

/**
 * Tests if we can start resizing a stencil
 */
bool SelectTool::startResizing(QMouseEvent *e)
{
    KivioStencil *pStencil;
    TKPoint pagePoint = m_pCanvas->mapFromScreen(e->pos());
    float x, y;

    x = pagePoint.x;
    y = pagePoint.y;

    // Search selected stencils to see if we have a resizing point
    pStencil = m_pCanvas->activePage()->selectedStencils()->first();
    while( pStencil )
    {
        m_resizeHandle = isOverResizeHandle(pStencil, x, y);
        if( m_resizeHandle > 0 )
        {
            m_pResizingStencil = pStencil;
//            m_pCanvas->startSelectedStencilResize( pStencil );

            // Create a new painter object
            m_pCanvas->beginUnclippedSpawnerPainter();
            m_pCanvas->drawStencilXOR( pStencil );

            return true;
        }

        pStencil = m_pCanvas->activePage()->selectedStencils()->next();
    }

    return false;
}



void SelectTool::mouseMove( QMouseEvent * e )
{
    switch( m_mode )
    {
        case stmDrawRubber:
            continueRubberBanding(e);
            break;

        case stmDragging:
            continueDragging(e);
            break;

        case stmCustomDragging:
            continueCustomDragging(e);
            break;

        case stmResizing:
            continueResizing(e);
            break;

        default:
            changeMouseCursor(e);
            break;
    }

    m_lastPoint = m_pCanvas->mapFromScreen( e->pos() );
}

void SelectTool::continueRubberBanding( QMouseEvent *e )
{
    m_pCanvas->continueRectDraw( e->pos(), KivioCanvas::Rubber );
}


void SelectTool::continueDragging( QMouseEvent *e )
{
    TKPoint pagePoint;
    KivioPoint kPoint;
    float dx, dy;

    pagePoint = m_pCanvas->mapFromScreen( e->pos() );
    dx = pagePoint.x - m_lastPoint.x;
    dy = pagePoint.y - m_lastPoint.y;

    // Undraw the old stencils
    m_pCanvas->drawSelectedStencilsXOR();


    // Translate to the new position
    KivioStencil *pStencil = m_pCanvas->activePage()->selectedStencils()->first();
    while( pStencil )
    {
        pStencil->setPosition( pStencil->x() + dx, pStencil->y() + dy );

        pStencil = m_pCanvas->activePage()->selectedStencils()->next();
    }

    // Draw the stencils
    m_pCanvas->drawSelectedStencilsXOR();
    m_pView->updateToolBars();
}

void SelectTool::continueCustomDragging( QMouseEvent *e )
{
    KivioCustomDragData data;
    TKPoint pagePoint;
    KivioPoint kPoint;

    pagePoint = m_pCanvas->mapFromScreen( e->pos() );

    data.page = m_pCanvas->activePage();
    data.dx = pagePoint.x - m_lastPoint.x;
    data.dy = pagePoint.y - m_lastPoint.y;
    data.x = pagePoint.x;
    data.y = pagePoint.y;
    data.id = m_customDragID;
    data.scale = m_pCanvas->view()->zoom() / 100.0f;


    // Undraw the old stencils
    m_pCanvas->drawSelectedStencilsXOR();


    // Translate to the new position
//    KivioStencil *pStencil = m_pCanvas->activePage()->selectedStencils()->first();
//    while( pStencil )
//    {
//        pStencil->customDrag( &data );
//
//        pStencil = m_pCanvas->activePage()->selectedStencils()->next();
//    }

    // Custom dragging can only occur on one stencil
    if( m_pCustomDraggingStencil )
        m_pCustomDraggingStencil->customDrag( &data );

    // Draw the stencils
    m_pCanvas->drawSelectedStencilsXOR();
    m_pView->updateToolBars();
}


void SelectTool::continueResizing( QMouseEvent *e )
{
    TKPoint pagePoint;
    KivioPoint kPoint;
    float dx, dy;

    pagePoint = m_pCanvas->mapFromScreen( e->pos() );
    dx = pagePoint.x - m_lastPoint.x;
    dy = pagePoint.y - m_lastPoint.y;

    // Undraw the old outline
    m_pCanvas->drawStencilXOR( m_pResizingStencil );


    float sx = m_pResizingStencil->x();
    float sy = m_pResizingStencil->y();
    float sw = m_pResizingStencil->w();
    float sh = m_pResizingStencil->h();

    switch( m_resizeHandle )
    {
        case 1: // top left
            if( m_pResizingStencil->protection()->testBit( kpWidth )==false )
            {
                m_pResizingStencil->setX( sx + dx );
                m_pResizingStencil->setW( sw - dx );
            }
                if( m_pResizingStencil->protection()->testBit( kpHeight )==false )
            {
                m_pResizingStencil->setY( sy + dy );
                m_pResizingStencil->setH( sh - dy );
            }
            break;

        case 2: // top
            if( m_pResizingStencil->protection()->testBit( kpHeight )==false )
            {
                m_pResizingStencil->setY( sy + dy );
                m_pResizingStencil->setH( sh - dy );
            }
            break;

        case 3: // top right
            if( m_pResizingStencil->protection()->testBit( kpHeight )==false )
           {
               m_pResizingStencil->setY( sy + dy );
               m_pResizingStencil->setH( sh - dy );
           }

           if( m_pResizingStencil->protection()->testBit( kpWidth )==false )
           {
               m_pResizingStencil->setW( sw + dx );
           }
           break;

        case 4: // right
            if( m_pResizingStencil->protection()->testBit( kpWidth )==false )
            {
                // see old kivio source when snaptogrid gets implemented
                //setX( SnapToGrid(sx+sw+dx)-sx )
                m_pResizingStencil->setW( sw + dx );
            }
            break;

        case 5: // bottom right
            if( m_pResizingStencil->protection()->testBit( kpWidth )==false )
            {
                m_pResizingStencil->setW( sw+dx );
            }

            if( m_pResizingStencil->protection()->testBit( kpHeight )==false )
            {
                m_pResizingStencil->setH( sh + dy );
            }
            break;

        case 6: // bottom
            if( m_pResizingStencil->protection()->testBit( kpHeight )==false )
            {
                m_pResizingStencil->setH( sh + dy );
            }
            break;

        case 7: // bottom left
            if( m_pResizingStencil->protection()->testBit( kpWidth )==false )
            {
                 m_pResizingStencil->setX( sx + dx );
                m_pResizingStencil->setW( sw - dx );
            }

            if( m_pResizingStencil->protection()->testBit( kpHeight )==false )
            {
                m_pResizingStencil->setH( sh + dy );
            }
            break;

        case 8: // left
            if( m_pResizingStencil->protection()->testBit( kpWidth )==false )
            {
                m_pResizingStencil->setX( sx + dx );
                m_pResizingStencil->setW( sw - dx );
            }
             break;

        default:
            debug("SelectTool::continueResizing() - unknown resize handle: %d", m_resizeHandle );
            break;
    }

    m_pCanvas->drawStencilXOR( m_pResizingStencil );
    m_pView->updateToolBars();
}


/**
 * Change the mouse cursor based on what it is over.
 */
void SelectTool::changeMouseCursor( QMouseEvent *e )
{
    TKPoint pagePoint = m_pCanvas->mapFromScreen( e->pos() );
    KivioStencil *pStencil;
    int cursorType;

    float x = pagePoint.x;
    float y = pagePoint.y;

    // Iterate through all the selected stencils
    pStencil = m_pCanvas->activePage()->selectedStencils()->first();
    while( pStencil )
    {
        cursorType = isOverResizeHandle(pStencil, x, y);
        switch( cursorType )
        {
            case 1: // top left
                m_pCanvas->setCursor( sizeFDiagCursor );
                return;

            case 2: // top
                m_pCanvas->setCursor( sizeVerCursor );
                return;

            case 3: // top right
                m_pCanvas->setCursor( sizeBDiagCursor );
                return;

            case 4: // right
                m_pCanvas->setCursor( sizeHorCursor );
                return;

            case 5: // bottom right
                m_pCanvas->setCursor( sizeFDiagCursor );
                return;

            case 6: // bottom
                m_pCanvas->setCursor( sizeVerCursor );
                return;

            case 7: // bottom left
                m_pCanvas->setCursor( sizeBDiagCursor );
                return;

            case 8: // left
                m_pCanvas->setCursor( sizeHorCursor );
                return;

            default:
                KivioPoint col;

                col.set(x,y);
                if( pStencil->checkForCollision( &col )==true )
                {
                    m_pCanvas->setCursor( sizeAllCursor );
                    return;
                }
                break;

        }


        pStencil = m_pCanvas->activePage()->selectedStencils()->next();
    }

    m_pCanvas->setCursor( arrowCursor );
}


/**
 * Tests a box for a point.  Used only by isOverResizeHandle().
 */
#define RESIZE_BOX_TEST( x, y, bx, by ) \
x >= bx-three_pixels && \
x <= bx+three_pixels && \
y >= by-three_pixels && \
y <= by+three_pixels

/**
 * Tests if a point is over a stencils
 */
int SelectTool::isOverResizeHandle( KivioStencil *pStencil, float x, float y )
{
    float newX, newY, newW, newH;

    newX = pStencil->x();
    newY = pStencil->y();
    newW = pStencil->w();
    newH = pStencil->h();

    float three_pixels = 4.0f;

    // Top left
    if( RESIZE_BOX_TEST( x, y, newX, newY ) )
        return 1;

    // Top
    if( RESIZE_BOX_TEST( x, y, newX+newW/2, newY ) )
        return 2;

    // Top right
    if( RESIZE_BOX_TEST( x, y, newX+newW, newY  ) )
        return 3;

    // Right
    if( RESIZE_BOX_TEST( x, y, newX+newW, newY+newH/2 ) )
        return 4;

    // Bottom right
    if( RESIZE_BOX_TEST( x, y, newX+newW, newY+newH ) )
        return 5;

    // Bottom
    if( RESIZE_BOX_TEST( x, y, newX+newW/2, newY+newH ) )
        return 6;

    // Bottom left
    if( RESIZE_BOX_TEST( x, y, newX, newY+newH ) )
        return 7;

    // Left
    if( RESIZE_BOX_TEST( x, y, newX, newY+newH/2 ) )
        return 8;

    // Nothing found
    return 0;
}


void SelectTool::mouseRelease( QMouseEvent *e )
{
    m_releasePoint = e->pos();

    switch( m_mode )
    {
        case stmDrawRubber:
            endRubberBanding(e);
            break;

        case stmCustomDragging:
            endCustomDragging(e);
            break;

        case stmDragging:
            endDragging(e);
            break;

        case stmResizing:
            endResizing(e);
            break;
    }
	
	m_mode = stmNone;

    m_pCanvas->repaint();
}

void SelectTool::endRubberBanding(QMouseEvent *e)
{
    // End the rubber-band drawing
    m_pCanvas->endRectDraw();

    // We can't select if the start and end points are the same
    if( m_startPoint != m_releasePoint )
    {
        select(m_pCanvas->rect());
    }

    m_pView->updateToolBars();
}

void SelectTool::endDragging(QMouseEvent *e)
{
    m_pCanvas->drawSelectedStencilsXOR();

    m_pCanvas->endUnclippedSpawnerPainter();
}

void SelectTool::endCustomDragging(QMouseEvent *e)
{
    m_customDragID = 0;
    m_pCanvas->drawSelectedStencilsXOR();

    m_pCanvas->endUnclippedSpawnerPainter();
}

void SelectTool::endResizing(QMouseEvent *e)
{
    // Undraw the last outline
    m_pCanvas->drawStencilXOR( m_pResizingStencil );

    // Deallocate the painter object
    m_pCanvas->endUnclippedSpawnerPainter();

    // Set the class vars to nothing
    m_pResizingStencil = NULL;
    m_resizeHandle = 0;
}
